/*****************************************************************
 * ew2ringserver.c
 *
 * Program to read (selected) messages from a transport ring, create
 * miniSEED and forward packets to a ringserver.  The source code
 * started as ringdup/tbuf2mseed as of 2013/6/1, credit for the kernel
 * used to start goes to those authors.
 *
 * ew2ringserver can listen for TRACEBUF[2] and MSEED message types.
 * For TRACEBUF[2] messages, the data are buffered and packed into
 * 512-byte miniSEED records.  For MSEED message types if the detected
 * record length is 512-bytes it is forwarded to the ringserver
 * unmodified, otherwise it is ignored.
 *
 * Hacked on by Chad Trabant among others.
 *
 *****************************************************************/

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <math.h>
#include <pthread.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "earthworm.h"
#include "imp_exp_gen.h"
#include "kom.h"
#include "libdali.h"
#include "libmseed2.h"
#include "mem_circ_queue.h"
#include "swap.h"
#include "time_ew.h"
#include "trace_buf.h"
#include "transport.h"

#include "exportfilter.h"

#define PACKAGE "ew2ringserver"
#define PROGRAM_VERSION "2.7 2024.10.28"

static thr_ret
HeartBeater (void *);                   /* Send periodic heartbeat messages to transport ring */
static thr_ret MessageStacker (void *); /* Read messages and add to queue */

static void config (char *);
static void lookup (void);
static void logstatus (unsigned char, short, char *);
static void freelocal (void);

static MSTraceGroup *mstg = 0; /* buffer for miniseed packaging */
static void sendrecord (char *record, int reclen, void *handlerdata);
static int handleqmsg (MSG_LOGO *Logo, char *msg, int nbytes);
static int handletbuf (char *msgbuf, int nbytes);
static int packtrace (MSTrace *mst, int flush);
static int packalltraces (int flush, hptime_t flushtime);
static void logmststats (MSTrace *mst);
static void clearmststats (MSTrace *mst);
static void logit_msg (const char *);
static void logit_err (const char *);

#ifndef WIN32
static void term_handler (int sig);
#endif

/* Per-trace statistics */
typedef struct tracestats_s
{
  hptime_t earliest;
  hptime_t latest;
  hptime_t update;
  hptime_t xmit;
  int64_t pktcount;
  int64_t reccount;
} TraceStats;

/* Thread things */
#define THREAD_STACK 8192
static ew_thread_t
    tidStacker; /* Thread moving messages from transport to queue */
static ew_thread_t
    tidHeartBeater; /* Thread to send heartbeat to transport ring */

#define MSGSTK_OFF 0   /* MessageStacker has not been started      */
#define MSGSTK_ALIVE 1 /* MessageStacker alive and well            */
#define MSGSTK_ERR -1  /* MessageStacker encountered error quit    */
volatile int MessageStackerStatus = MSGSTK_OFF;

QUEUE MsgQueue; /* from queue.h, queue.c; sets up linked    */
static pthread_mutex_t mutex =
    PTHREAD_MUTEX_INITIALIZER; /* to lock/unlock queue */
static pthread_cond_t condLess =
    PTHREAD_COND_INITIALIZER; /* when queue size decreases */
static pthread_cond_t condMore =
    PTHREAD_COND_INITIALIZER; /* when queue size increases */

/* Message Buffers to be allocated */
static char *Rawmsg      = NULL; /* "raw" retrieved message                  */
static char *Filteredmsg = NULL; /* MessageStacker's "filtered" message      */
static char *Qmsg        = NULL; /* message from queue                       */
static MSG_LOGO Logo;            /* logo of message to re-send               */

static SHM_INFO InRegion; /* shared memory region to use for input  */

#define MAXLOGO 10
MSG_LOGO GetLogo[MAXLOGO]; /* array for requesting module,type,instid */
short nLogo;

char *Argv0; /* pointer to executable name */
pid_t MyPid; /* Our own pid, sent with heartbeat for restart purposes */

/* Things to read or derive from configuration file */
static char RingName[MAX_RING_STR]; /* name of transport ring for input  */
static char MyModName[MAX_MOD_STR]; /* speak as this module name/id      */
static int LogSwitch;               /* 0 if no logfile should be written */
static int HeartBeatInt;            /* seconds between heartbeats        */
static int EmptyQWait = -1;         /* Empty queue wait time for MsgStacker
                                       in seconds (default is HeartBeatInt) */
static int FullQWait = -1;          /* Full queue wait time for MsgStacker
                                       in seconds (default is HeartBeatInt * 3) */
static long MaxMsgSize;             /* max size for input/output msgs    */
static int QueueSize;               /* max messages in output circular buffer */

/* Things to look up in the earthworm.h tables with getutil.c functions */
static long RingNameKey;      /* key of transport ring for input    */
static unsigned char InstId;  /* local installation id              */
static unsigned char MyModId; /* Module Id for this program         */
static unsigned char TypeHeartBeat;
static unsigned char TypeError;
static unsigned char TypeMseed;
static unsigned char TypeTracebuf;
static unsigned char TypeTracebuf2;

static int msreclen = 512;

/* A non-official data encoding for "Steim-2 then Steim-1 then 32-bit integers"
 * logic */
#define DYNAMIC_STEIM21I -11
static int int32encoding =
    DYNAMIC_STEIM21I; /* Data sample encoding for 32-bit integer data */

/* Error messages used by export */
#define ERR_MISSMSG 0      /* message missed in transport ring        */
#define ERR_TOOBIG 1       /* retreived msg too large for buffer      */
#define ERR_NOTRACK 2      /* msg retreived; tracking limit exceeded  */
#define ERR_QUEUE 3        /* error queueing message for sending      */
static char errText[1024]; /* string for log/error messages           */

static volatile int stopsig =
    0; /* 1: termination requested, 2: termination and no flush */
static int verbose = 0;
static int reconnectinterval =
    10; /* Interval to wait between reconnection attempts in seconds */
static int flushlatency =
    300; /* Flush data buffers if not updated for latency in seconds */

static int flushfastrate =
    0; /* Flush data buffers if sample rate >= this value, 0: disabled */
static hptime_t flushfastduration =
    0; /* Desired miniSEED duration in HPTMODULUS, 0: disabled */
static int dequeueLoopCount =
    10; /* # of attempts by producer thread  to get msgs from queue */

static char *rsaddr =
    0;                 /* DataLink/ringserver receiver address in IP:port format */
static DLCP *dlcp = 0; /* DataLink connection handle */

int
main (int argc, char **argv)
{
  int res;
  long recsize;     /* size of retrieved message             */
  MSG_LOGO reclogo; /* logo of retrieved message             */

  int ret;
  int rc;
  long msgSize;
  int count;

  time_t now;         /* current time, used for timing buffer flushes */
  time_t MyLastFlush; /* time of last buffer flush */

#ifndef WIN32
  /* Signal handling, use POSIX calls with standardized semantics */
  struct sigaction sa;

  sa.sa_flags = SA_RESTART;
  sigemptyset (&sa.sa_mask);

  sa.sa_handler = term_handler;
  sigaction (SIGINT, &sa, NULL);
  sigaction (SIGQUIT, &sa, NULL);
  sigaction (SIGTERM, &sa, NULL);

  sa.sa_handler = SIG_IGN;
  sigaction (SIGHUP, &sa, NULL);
  sigaction (SIGPIPE, &sa, NULL);
#endif

  /* Check command line arguments */
  Argv0 = argv[0];
  if (argc != 2)
  {
    fprintf (stderr, "Usage: %s <configfile>\n", Argv0);
    fprintf (stderr, "Version: V%s\n", PROGRAM_VERSION);
    exit (0);
  }

  /* Initialize name of log-file & open it */
  logit_init (argv[1], 0, 512, 1);

  /* Read the configuration file(s) */
  config (argv[1]);
  logit ("et", "%s(%s): Read command file <%s>\n Version %s\n", Argv0, MyModName,
         argv[1], PROGRAM_VERSION);

  /* Look up important info from earthworm.h tables */
  lookup ();

  /* Initialize trace buffer */
  if (!(mstg = mst_initgroup (mstg)))
  {
    logit ("et", "%s(): Cannot initialize MSTraceList\n", __func__);
    exit (1);
  }

  /* Reinitialize the logging level */
  logit_init (argv[1], 0, 512, LogSwitch);

  /* Get our own Pid for restart purposes */
  MyPid = getpid ();
  if (MyPid == -1)
  {
    logit ("e", "%s(): Cannot get pid; exiting!\n", __func__);
    exit (0);
  }

  logit ("e", "PID=%d\n", (int)MyPid);

  /* Initialize export filter */
  if (exportfilter_init () != 0)
  {
    logit ("e", "%s(): Error in exportfilter_init(); exiting!\n", __func__);
    exit (-1);
  }

  /* Allocate space for input/output messages for all threads */
  if ((Rawmsg = (char *)malloc (MaxMsgSize + 1)) == NULL)
  {
    logit ("e", "%s(%s): error allocating Rawmsg; exiting!\n", Argv0, MyModName);
    freelocal ();
    exit (-1);
  }

  /* Buffers for message handling loop */
  if ((Qmsg = (char *)malloc (MaxMsgSize + 1)) == NULL)
  {
    logit ("e", "%s(): error allocating Qmsg; exiting!\n", __func__);
    freelocal ();
    exit (-1);
  }

  /* Buffers for the MessageStacker thread: */
  if ((Filteredmsg = (char *)malloc (MaxMsgSize + 1)) == NULL)
  {
    logit ("e", "%s(): error allocating Filteredmsg; exiting!\n", __func__);
    freelocal ();
    exit (-1);
  }

  /* Create a Mutex to control access to queue */
  CreateSpecificMutex (&mutex);

  /* Initialize the message queue */
  initqueue (&MsgQueue, (unsigned long)QueueSize, (unsigned long)MaxMsgSize + 1);

  /* Allocate and initialize DataLink connection description */
  if (!(dlcp = dl_newdlcp (rsaddr, PACKAGE)))
  {
    logit ("e", "%s(): Cannot allocation DataLink descriptor\n", __func__);
    exit (-1);
  }

  /* Connect to destination DataLink server */
  if (dl_connect (dlcp) < 0)
  {
    logit ("t",
           "%s(): Initial connection to DataLink server (%s) failed, will retry "
           "later\n",
           __func__, dlcp->addr);
    dl_disconnect (dlcp);
  }
  else
    logit ("t", "%s(): Connected to ringserver at %s\n", dlcp->addr, __func__);

  /* Attach to Input shared memory ring */
  tport_attach (&InRegion, RingNameKey);

  /* step over all messages from transport ring */
  /* As Lynn pointed out: if we're restarted by startstop after hanging,
     we should throw away any of our messages in the transport ring.
     Else we could end up re-sending a previously sent message, causing
     time to go backwards... */
  do
  {
    res = tport_getmsg (&InRegion, GetLogo, nLogo, &reclogo, &recsize, Rawmsg,
                        MaxMsgSize);
  } while (res != GET_NONE);

  /* Start heartbeat thread */
  if (StartThread (HeartBeater, (unsigned)THREAD_STACK, &tidHeartBeater) == -1)
  {
    logit ("e", "%s(): Error starting heart beat thread; exiting!\n", __func__);
    tport_detach (&InRegion);
    freelocal ();
    exit (-1);
  }

  /* Start the message stacking thread if it isn't already running. */
  if (MessageStackerStatus != MSGSTK_ALIVE)
  {
    if (StartThread (MessageStacker, (unsigned)THREAD_STACK, &tidStacker) ==
        -1)
    {
      logit ("e", "%s(): Error starting MessageStacker thread; exiting!\n",
             __func__);
      tport_detach (&InRegion);
      freelocal ();
      exit (-1);
    }
    MessageStackerStatus = MSGSTK_ALIVE;
  }

  /* Start main service loop */
  while (1)
  {
    rc = 0;

    if (stopsig != 0)
    {
      logit ("et", "%s(): stopsig=%d\n", __func__, stopsig);
      break;
    }

    rc = tport_getflag (&InRegion);

    if (rc == TERMINATE)
    {
      logit ("et", "%s(): tport TERMINATE flag set, exiting while loop\n",
             __func__);
      break;
    }

    if (rc == MyPid)
    {
      logit ("et",
             "%s(): tport restart flag set for MyPid=%d, exiting while loop\n",
             __func__, MyPid);
      break;
    }

    if (MessageStackerStatus != MSGSTK_ALIVE)
    {
      logit ("et", "%s(): MessageStackerStatus=%d, exiting main while loop\n",
             __func__, MessageStackerStatus);
      break;
    }

    time (&now);
    /* Check if idle data streams need flushing */
    if (difftime (now, MyLastFlush) > (double)flushlatency)
    {
      if (packalltraces (0, DL_EPOCH2DLTIME (now) -
                                DL_EPOCH2DLTIME (flushlatency)) < 0)
      {
        logit ("et", "%s(): Cannot pack idle trace buffers or send records!\n",
               __func__);
      }
      MyLastFlush = now;
    }

    /* Process up to 10 messages, then let outer loop have a pass */
    count = 0;
    while (count < dequeueLoopCount)
    {
      /* Get message from queue */
      RequestSpecificMutex (&mutex);
      if (MsgQueue.NumOfElements <= 0)
      { /* empty queue */

        // wait for queue size increase in MessageStacker
        if (EmptyQWait) // if timeout
        {
          struct timespec ts;
          clock_gettime (CLOCK_REALTIME, &ts);
          ts.tv_sec += (time_t)(EmptyQWait);
          rc = pthread_cond_timedwait (&condMore, &mutex, &ts);
        }
        else
        {
          rc = pthread_cond_wait (&condMore, &mutex);
        }

        if (rc != 0)
        {
          if (rc != ETIMEDOUT)
          {
            logit ("et",
                   "Error: %s(): main thread: %s: %s\n",
                   __func__, EmptyQWait ? "pthread_cond_timedwait" : "pthread_cond_wait", strerror (rc));
            exit (-1); /* serious error */
          }
          else
          {
            logit ("et",
                   "Warning: %s(): main thread: %s: %s\n",
                   __func__, EmptyQWait ? "pthread_cond_timedwait" : "pthread_cond_wait", strerror (rc));
            if (MsgQueue.NumOfElements <= 0)
            {
              sprintf (
                  errText,
                  "Warning: %s(%s): Queue is empty, waiting for producer ...",
                  Argv0, MyModName);
              logstatus (TypeError, ERR_MISSMSG, errText);
            }
          }
        }
      }

      if (MsgQueue.NumOfElements > 0)
      {
        ret = dequeue (&MsgQueue, Qmsg, &msgSize, &Logo);
        assert (ret == 0);

        // signal queue size decrease to MessageStacker
        rc = pthread_cond_signal (&condLess);
        if (rc != 0)
        {
          logit (
              "et",
              "Error:  %s(): main thread: pthread_cond_signal error code: %s\n",
              __func__, strerror (rc));
          exit (-1);
        }
      }

      ReleaseSpecificMutex (&mutex);

      handleqmsg (&Logo, Qmsg, msgSize);

      count++;
    }
  } /* End of main service loop */

  /* Shut it all down */
  while (1) /* Flush message queue */
  {
    RequestSpecificMutex (&mutex);
    ret = dequeue (&MsgQueue, Qmsg, &msgSize, &Logo);
    ReleaseSpecificMutex (&mutex);

    if (ret < 0)
      break;

    handleqmsg (&Logo, Qmsg, msgSize);
  }

  CloseSpecificMutex (&mutex);

  stopsig = 1;

  tport_detach (&InRegion);

  packalltraces (1, HPTERROR); /* Flush all data in pack buffers */

  if (dlcp->link != -1)
    dl_disconnect (dlcp);

  exportfilter_shutdown ();

  freelocal ();

  if (verbose)
  {
    MSTrace *mst = mstg->traces;

    while (mst)
    {
      logmststats (mst);
      mst = mst->next;
    }
  }

  logit ("t", "%s(): termination requested; exiting!\n", __func__);

  return 0;
} /* End of main() */

/********************** Heartbeat Thread *******************
 *            Beat the heart into the transport ring              *
 ******************************************************************/
static thr_ret
HeartBeater (void *dummy)
{
  time_t now;        /* current time, used for timing heartbeats */
  time_t MyLastBeat; /* time of last local (into Earthworm) hearbeat */

  time (&MyLastBeat);

  while (stopsig == 0)
  {
    time (&now);
    if (difftime (now, MyLastBeat) > (double)HeartBeatInt)
    {
      if (verbose >= 2)
      {
        logit ("et",
               "Warning: %s() Heartbeat interval exceeded, issuing heartbeat!\n",
               __func__);
      }
      logstatus (TypeHeartBeat, 0, "");
      MyLastBeat = now;
    }

    sleep_ew (1000); /* to not hog CPU */
  }

  /* Ugly!? But EW lib forces all threads to be detached! */
  sleep_ew (5000); // wait for main thread to shut down
  logit ("et",
         "%s(): main thread did not terminate, exiting from heartbeat thread\n",
         __func__);
  KillSelfThread ();

  return THR_NULL_RET;
}

/********************** Message Stacking Thread *******************
 *           Move messages from transport to memory queue         *
 ******************************************************************/
static thr_ret
MessageStacker (void *dummy)
{
  long recsize;               /* size of retrieved message             */
  MSG_LOGO reclogo;           /* logo of retrieved message             */
  long filteredSize;          /* size of message after user-filtering  */
  unsigned char filteredType; /* type of message after filtering   */
  int res, ret, rc;
  int NumOfTimesQueueLapped = 0; /* number of messages lost due to
                                       queue lap */
  float pc_full;

  /* Tell the main thread we're ok
  ********************************/
  MessageStackerStatus = MSGSTK_ALIVE;

  /* Start main export service loop for current connection
  ********************************************************/
  while (1)
  {
    /* Get a message from transport ring
    ************************************/
    res = tport_getmsg (&InRegion, GetLogo, nLogo, &reclogo, &recsize, Rawmsg,
                        MaxMsgSize);

    /* Wait if no messages for us
     ****************************/
    if (res == GET_NONE)
    {
      sleep_ew (100);
      continue;
    }

    /* ignore msg if too big */
    if (res == GET_TOOBIG)
    {
      sprintf (errText, "(unfiltered) msg[%ld] i%d m%d t%d too long for target",
               recsize, (int)reclogo.instid, (int)reclogo.mod,
               (int)reclogo.type);
      logstatus (TypeError, ERR_TOOBIG, errText);
      continue;
    }

    /* pass it through the filter routine: this may reformat,
       or reject as it chooses. */
    Rawmsg[recsize] = '\0';
    if (exportfilter (Rawmsg, recsize, reclogo.type, &Filteredmsg, &filteredSize,
                      &filteredType) == 0)
    {
      continue;
    }

    pc_full = (float)MsgQueue.NumOfElements / QueueSize;
    if (pc_full > 90.0)
    {
      logit ("et",
             "Warning: %s(): MsgStacker thread: Queue %.2f full, turning on "
             "verbose mode\n",
             __func__, pc_full);
      verbose = 1;
      dl_loginit (verbose, &logit_msg, NULL, &logit_err, NULL);
    }

    /* Check return code; report errors
    ***********************************/
    if (res != GET_OK)
    {
      TRACE2_HEADER *h = (TRACE2_HEADER *)Rawmsg;
      if (res == GET_MISS)
      {
        sprintf (errText,
                 "Warning: %s(%s) missed msg(s) i%d m%d t%d in %s : %.4f "
                 "%s.%s.%s.%s",
                 Argv0, MyModName, (int)reclogo.instid, (int)reclogo.mod,
                 (int)reclogo.type, RingName, h->starttime, h->sta, h->chan,
                 h->net, h->loc);
        logstatus (TypeError, ERR_MISSMSG, errText);

        /* quit if missed msgs and queue is full */
        if (MsgQueue.NumOfElements >= QueueSize)
        {
          sprintf (errText,
                   "Error: %s(%s):MsgStacker thread: Queue is full + missed "
                   "msgs. Terminating ...",
                   Argv0, MyModName);
          logstatus (TypeError, ERR_MISSMSG, errText);
          break;
        }
      }
      else if (res == GET_NOTRACK)
      {
        sprintf (errText,
                 "no tracking for logo i%d m%d t%d in %s : %.4f %s.%s.%s.%s",
                 (int)reclogo.instid, (int)reclogo.mod, (int)reclogo.type,
                 RingName, h->starttime, h->sta, h->chan, h->net, h->loc);
        logstatus (TypeError, ERR_NOTRACK, errText);
      }
    }

    reclogo.type = filteredType; /* note the new message type */

    /* Process retrieved msg (res==GET_OK,GET_MISS,GET_NOTRACK)
    ***********************************************************/

    /* put it into the 'to be shipped' queue */
    /* the main thread is in the biz of de-queueng and handling */
    RequestSpecificMutex (&mutex);

    if (MsgQueue.NumOfElements >= QueueSize)
    {
      // wait for queue size decrease in main thread
      if (FullQWait) // if timeout
      {
        struct timespec ts;
        clock_gettime (CLOCK_REALTIME, &ts);
        ts.tv_sec += (time_t)(FullQWait);
        rc = pthread_cond_timedwait (&condLess, &mutex, &ts);
      }
      else
      {
        rc = pthread_cond_wait (&condLess, &mutex);
      }

      if (rc != 0)
      {
        logit ("et",
               "Error:  %s(): MsgStacker thread: %s: %s\n",
               __func__, FullQWait ? "pthread_cond_timedwait" : "pthread_cond_wait", strerror (rc));
        if (rc == ETIMEDOUT && MsgQueue.NumOfElements >= QueueSize)
        {
          /* this could happen if main(consumer) thread does not keep up */
          sprintf (errText, "Warning: %s(%s): Queue is full ...", Argv0,
                   MyModName);
          logstatus (TypeError, ERR_MISSMSG, errText);
        }

        logit ("et",
               "Error:  %s(): MsgStacker thread: %s: %s, "
               "producer thread quitting\n",
               __func__, FullQWait ? "pthread_cond_timedwait" : "pthread_cond_wait", strerror (rc));
        ReleaseSpecificMutex (&mutex);

        break; /* exit while loop */
      }
    }

    assert (MsgQueue.NumOfElements < QueueSize);
    ret = enqueue (&MsgQueue, Filteredmsg, filteredSize, reclogo);
    assert (ret == 0 || ret == -1 || ret == -3);

    // signal queue size increase to main thread
    rc = pthread_cond_signal (&condMore);
    if (rc != 0)
    {
      logit ("et",
             "%s(): Error: MsgStacker thread: enqueue: pthread_cond_signal "
             "error code %s\n",
             __func__, strerror (rc));
      exit (-1);
    }

    ReleaseSpecificMutex (&mutex);

    switch (ret)
    { /* examine enqueue issues */
    case 0:
      if (NumOfTimesQueueLapped > 0)
      {
        NumOfTimesQueueLapped = 0;
      }
      break;
    case -1:
      sprintf (errText,
               "Message too big, queue cannot allocate memory. Lost message.");
      logstatus (TypeError, ERR_QUEUE, errText);
      break;
    case -3:
      NumOfTimesQueueLapped++;
      if (NumOfTimesQueueLapped == 1 || !(NumOfTimesQueueLapped % 5))
      {
        logit ("t",
               "%s(): Circular queue lapped %d time(s). Messages lost. Queue "
               "has %ld messages.\n",
               __func__, NumOfTimesQueueLapped, MsgQueue.NumOfElements);
      }
      break;
    default:
      break;
    }

    continue; /* get more data */

  } /* end of while */

  /* we're quitting
  *****************/
  MessageStackerStatus = MSGSTK_ERR; /* file a complaint to the main thread */
  KillSelfThread ();                 /* main thread will restart us */

  return THR_NULL_RET;
}

/*****************************************************************************
 *  config() processes command file(s) using kom.c functions;         *
 *                    exits if any errors are encountered.               *
 *****************************************************************************/
static void
config (char *configfile)
{
  int ncommand;  /* # of required commands you expect to process   */
  char init[10]; /* init flags, one byte for each required command */
  int nmiss;     /* number of required commands that were missed   */
  char *com;
  char processor[20];
  int nfiles;
  int success;
  int i;
  char *str;

  /* Set to zero one init flag for each required command */
  ncommand = 8;
  for (i = 0; i < ncommand; i++)
    init[i] = 0;
  nLogo = 0;

  /* Open the main configuration file */
  nfiles = k_open (configfile);
  if (nfiles == 0)
  {
    logit ("e", "%s(): Error opening command file <%s>; exiting!\n", __func__,
           configfile);
    exit (-1);
  }

  /* Process all command files */
  while (nfiles > 0) /* While there are command files open */
  {
    while (k_rd ()) /* Read next line from active file */
    {
      com = k_str (); /* Get the first token from line */

      /* Ignore blank lines & comments */
      if (!com)
        continue;
      if (com[0] == '#')
        continue;

      /* Open a nested configuration file */
      if (com[0] == '@')
      {
        success = nfiles + 1;
        nfiles  = k_open (&com[1]);
        if (nfiles != success)
        {
          logit ("e", "%s(): Error opening command file <%s>; exiting!\n",
                 __func__, &com[1]);
          exit (-1);
        }
        continue;
      }
      strcpy (processor, "config");

      /* Process anything else as a command */
      /*0*/ if (k_its ("LogFile"))
      {
        LogSwitch = k_int ();
        init[0]   = 1;
      }
      /*1*/ else if (k_its ("MyModuleId"))
      {
        str = k_str ();
        if (str)
          strcpy (MyModName, str);
        init[1] = 1;
      }
      /*2*/ else if (k_its ("RingName"))
      {
        str = k_str ();
        if (str)
          strcpy (RingName, str);
        init[2] = 1;
      }
      /*3*/ else if (k_its ("RSAddress"))
      {
        str = k_str ();
        if (str)
          rsaddr = strdup (str);
        init[3] = 1;
      }
      /*4*/ else if (k_its ("HeartBeatInt"))
      {
        HeartBeatInt = k_int ();
        init[4]      = 1;
      }

      /* Enter installation & module & message types to get */
      /*5*/ else if (k_its ("GetMsgLogo"))
      {
        if (nLogo >= MAXLOGO)
        {
          logit (
              "e",
              "%s(): Too many <GetMsgLogo> commands in <%s> max=%d; exiting!\n",
              __func__, configfile, (int)MAXLOGO);
          exit (-1);
        }
        if ((str = k_str ()))
        {
          if (GetInst (str, &GetLogo[nLogo].instid) != 0)
          {
            logit ("e",
                   "%s(): Invalid installation name <%s> in <GetMsgLogo> cmd; "
                   "exiting!\n",
                   __func__, str);
            exit (-1);
          }
        }
        if ((str = k_str ()))
        {
          if (GetModId (str, &GetLogo[nLogo].mod) != 0)
          {
            logit ("e",
                   "%s(): Invalid module name <%s> in <GetMsgLogo> cmd; "
                   "exiting!\n",
                   __func__, str);
            exit (-1);
          }
        }
        if ((str = k_str ()))
        {
          if (GetType (str, &GetLogo[nLogo].type) != 0)
          {
            logit ("e",
                   "%s(): Invalid msgtype <%s> in <GetMsgLogo> cmd; exiting!\n",
                   __func__, str);
            exit (-1);
          }
        }
        nLogo++;
        init[5] = 1;
      }

      /* Maximum size (bytes) for incoming/outgoing messages */
      /*6*/ else if (k_its ("MaxMsgSize"))
      {
        MaxMsgSize = k_long ();
        init[6]    = 1;
      }

      /* Maximum number of messages in outgoing circular buffer */
      /*7*/ else if (k_its ("QueueSize"))
      {
        QueueSize = k_long ();
        init[7]   = 1;
      }

      /* Optinal options */
      else if (k_its ("EmptyQWait"))
      {
        EmptyQWait = k_int ();
      }
      else if (k_its ("FullQWait"))
      {
        FullQWait = k_int ();
      }
      else if (k_its ("Verbosity"))
      {
        str = k_str ();
        if (str)
          verbose = atoi (str);
      }
      else if (k_its ("FlushLatency"))
      {
        str = k_str ();
        if (str)
          flushlatency = atoi (str);
      }
      else if (k_its ("ReconnectInterval"))
      {
        str = k_str ();
        if (str)
          reconnectinterval = atoi (str);
      }
      else if (k_its ("Int32Encoding"))
      {
        str = k_str ();
        if (str)
        {
          if (!strcmp (str, "STEIM1"))
            int32encoding = DE_STEIM1;
          else if (!strcmp (str, "STEIM2"))
            int32encoding = DE_STEIM2;
          else if (!strcmp (str, "STEIM21I"))
            int32encoding = DYNAMIC_STEIM21I;
          else if (!strcmp (str, "INT32"))
            int32encoding = DE_INT32;
          else
          {
            logit ("e", "%s(): Unrecognized Int32Encoding: %s; exiting!\n",
                   __func__, str);
            exit (-1);
          }
        }
      }
      else if (k_its ("FlushFastRate"))
      {
        str = k_str ();
        if (str)
          flushfastrate = atoi (str);
      }
      else if (k_its ("FlushFastDuration"))
      {
        str = k_str ();
        if (str)
          flushfastduration = (hptime_t)(atol (str));
      }
      else if (k_its ("DequeueLoopCount"))
      {
        str = k_str ();
        if (str)
          dequeueLoopCount = atoi (str);
      }
      /* Pass it off to the filter's config processor */
      else if (exportfilter_com ())
        strcpy (processor, "exportfilter_com");
      /* Unknown command */
      else
      {
        logit ("e", "%s(): <%s> Unknown command in <%s>.\n", __func__, com,
               configfile);
        continue;
      }

      /* See if there were any errors processing the command */
      if (k_err ())
      {
        logit ("e", "%s(): Bad <%s> command for %s() in <%s>; exiting!\n",
               __func__, com, processor, configfile);
        exit (-1);
      }
    }

    nfiles = k_close ();
  }

  /* After all files are closed, check init flags for missed commands */
  nmiss = 0;
  for (i = 0; i < ncommand; i++)
    if (!init[i])
      nmiss++;
  if (nmiss)
  {
    logit ("e", "%s(): ERROR, no ", __func__);
    if (!init[0])
      logit ("e", "<LogFile> ");
    if (!init[1])
      logit ("e", "<MyModuleId> ");
    if (!init[2])
      logit ("e", "<RingName> ");
    if (!init[3])
      logit ("e", "<RSAddress> ");
    if (!init[4])
      logit ("e", "<HeartBeatInt> ");
    if (!init[5])
      logit ("e", "<GetMsgLogo> ");
    if (!init[6])
      logit ("e", "<MaxMsgSize> ");
    if (!init[7])
      logit ("e", "<QueueSize> ");
    logit ("e", "command(s) in <%s>; exiting!\n", configfile);
    exit (-1);
  }

  if (EmptyQWait < 0)
  {
    EmptyQWait = HeartBeatInt;
  }
  if (FullQWait < 0)
  {
    FullQWait = HeartBeatInt * 3;
  }
  /* Redirect libmseed2 & libdali logging to EW facility through shim functions
   */
  ms_loginit (&logit_msg, NULL, &logit_err, NULL);
  dl_loginit (verbose, &logit_msg, NULL, &logit_err, NULL);

  return;
} /* End of config() */

/****************************************************************************
 *  lookup( )   Look up important info from earthworm.h tables       *
 ****************************************************************************/
static void
lookup (void)
{
  /* Look up keys to shared memory regions */
  if ((RingNameKey = GetKey (RingName)) == -1)
  {
    fprintf (stderr, "%s:  Invalid ring name <%s>; exiting!\n", Argv0, RingName);
    exit (-1);
  }

  /* Look up installations of interest */
  if (GetLocalInst (&InstId) != 0)
  {
    fprintf (stderr, "%s: error getting local installation id; exiting!\n",
             Argv0);
    exit (-1);
  }

  /* Look up modules of interest */
  if (GetModId (MyModName, &MyModId) != 0)
  {
    fprintf (stderr, "%s: Invalid module name <%s>; exiting!\n", Argv0,
             MyModName);
    exit (-1);
  }

  /* Look up message types of interest */
  if (GetType ("TYPE_HEARTBEAT", &TypeHeartBeat) != 0)
  {
    fprintf (stderr, "%s: Invalid message type <TYPE_HEARTBEAT>; exiting!\n",
             Argv0);
    exit (-1);
  }
  if (GetType ("TYPE_ERROR", &TypeError) != 0)
  {
    fprintf (stderr, "%s: Invalid message type <TYPE_ERROR>; exiting!\n", Argv0);
    exit (-1);
  }
  if (GetType ("TYPE_MSEED", &TypeMseed) != 0)
  {
    fprintf (stderr,
             "%s: Critical message type <TYPE_MSEED> is missing; exiting!\n",
             Argv0);
    exit (-1);
  }
  if (GetType ("TYPE_TRACEBUF", &TypeTracebuf) != 0)
  {
    fprintf (stderr,
             "%s: Critical message type <TYPE_TRACEBUF> is missing; exiting!\n",
             Argv0);
    exit (-1);
  }
  if (GetType ("TYPE_TRACEBUF2", &TypeTracebuf2) != 0)
  {
    fprintf (stderr,
             "%s: Critical message type <TYPE_TRACEBUF2> is missing; exiting!\n",
             Argv0);
    exit (-1);
  }

  return;
}

/***************************************************************************
 * logstatus() builds a heartbeat or error message & puts it into      *
 *                 shared memory.  Writes errors to log file & screen.     *
 ***************************************************************************/
static void
logstatus (unsigned char type, short ierr, char *note)
{
  MSG_LOGO logo;
  char msg[256];
  long size;
  time_t t;

  /* Build the message */
  logo.instid = InstId;
  logo.mod    = MyModId;
  logo.type   = type;

  time (&t);

  if (type == TypeHeartBeat)
    sprintf (msg, "%ld %ld\n", (long)t, (long)MyPid);
  else if (type == TypeError)
  {
    sprintf (msg, "%ld %hd %s\n", (long)t, ierr, note);

    logit ("et", "%s(): %s\n", __func__, note);
  }

  size = strlen (msg); /* don't include the null byte in the message */

  /* Write the message to shared memory */
  if (tport_putmsg (&InRegion, &logo, size, msg) != PUT_OK)
  {
    if (type == TypeHeartBeat)
    {
      logit ("et", "%s(): %s Error sending heartbeat.\n", __func__, MyModName);
    }
    else if (type == TypeError)
    {
      logit ("et", "%s(): %s Error sending error:%d.\n", __func__, MyModName,
             ierr);
    }
  }

  return;
}

/***************************************************************************
 * freelocal()  free all previously allocated memory                      *
 ***************************************************************************/
static void
freelocal (void)
{
  free (Rawmsg);      /* "raw" retrieved message   */
  free (Qmsg);        /* Incoming message buffer   */
  free (Filteredmsg); /* MessageStacker's "filtered" message to  */
                      /*  be sent to destination                 */
  return;
}

/*********************************************************************
 * handleqmsg:
 *
 * Handle message from queue, using different paths depending on type.
 *
 * Returns the number of records packed/transmitted on success and -1
 * on error.
 *********************************************************************/
static int
handleqmsg (MSG_LOGO *Logo, char *msg, int nbytes)
{
  /* Send TRACEBUF/TRACEBUF2 to buffer for packing */
  if (Logo->type == TypeTracebuf || Logo->type == TypeTracebuf2)
  {
    return handletbuf (Qmsg, nbytes);
  }
  /* Send 512-byte miniSEED record message directly to ringserver */
  else if (Logo->type == TypeMseed)
  {
    if (ms_detect (Qmsg, nbytes) == 512)
    {
      sendrecord (Qmsg, 512, NULL);
      return 1;
    }

    return 0;
  }
  else if (verbose)
  {
    logit ("t", "%s(): Unrecognized message type (%d), skipping\n", __func__,
           Logo->type);
  }
  return -1;
} /* End of handleqmsg() */

/*********************************************************************
 * handletbuf:
 *
 * Add the tbuf packet data to the trace buffer, pack the buffer and
 * send it on.
 *
 * Returns the number of records packed/transmitted on success and -1
 * on error.
 *********************************************************************/
static int
handletbuf (char *msg, int nbytes)
{
  static MSRecord *msr  = NULL;
  MSTrace *mst          = NULL;
  int recordspacked     = 0;
  MSTrace *prevmst      = NULL;
  MSTrace *existing_mst = NULL;
  hptime_t mst_duration = 0;

  static int32_t *int32buffer = NULL;
  static int32_t int32count   = 0;
  int16_t *int16buffer        = NULL;

  int idx;
  int rv = 0;
  TRACE2_HEADER *wfhead; /* pntr to header of TYPE_TRACEBUF2 msg */
  TracePacket *tp;
  char originaldatatype[3];
  int flushflag = 0;

  if (!msg || nbytes <= 0)
    return -1;

  tp     = (TracePacket *)msg;
  wfhead = &(tp->trh2);
  strncpy (originaldatatype, wfhead->datatype, sizeof (originaldatatype));

  rv = WaveMsg2MakeLocal (wfhead);
  if (rv != 0)
  {
    logit ("et", "%s(): Error with WaveMsg2MakeLocal(): %d\n", __func__, rv);
    return -1;
  }

  if (verbose > 1)
    logit ("et",
           "%s(): Received: %s.%s.%s.%s  ts=%.4f ns=%d  dt=%4.1f  type=%s (now "
           "%s)\n",
           __func__, wfhead->sta, wfhead->chan, wfhead->net, wfhead->loc,
           wfhead->starttime, wfhead->nsamp, wfhead->samprate, originaldatatype,
           wfhead->datatype);

  if (!(msr = msr_init (msr)))
  {
    logit ("et", "%s(): Could not (re)initialize MSRecord\n", __func__);
    return -1;
  }

  /* Populate an MSRecord from a tbuf2 */
  ms_strncpclean (msr->network, wfhead->net, 2);
  ms_strncpclean (msr->station, wfhead->sta, 5);
  if (strcmp (wfhead->loc, "--") == 0)
  {
    wfhead->loc[0] = ' ';
    wfhead->loc[1] = ' ';
  }
  ms_strncpclean (msr->location, wfhead->loc, 2);
  ms_strncpclean (msr->channel, wfhead->chan, 3);

  msr->starttime = (hptime_t)(MS_EPOCH2HPTIME (wfhead->starttime) + 0.5);
  msr->samprate  = wfhead->samprate;

  msr->datasamples = msg + sizeof (TRACE2_HEADER);
  msr->numsamples  = wfhead->nsamp;
  msr->samplecnt   = wfhead->nsamp;

  if (!strcmp (wfhead->datatype, "s2") || !strcmp (wfhead->datatype, "i2"))
  {
    /* (Re)Allocate static conversion buffer if more space is needed */
    if (int32count < msr->numsamples)
    {
      if ((int32buffer = (int32_t *)realloc (
               int32buffer, sizeof (int32_t) * msr->numsamples)) == NULL)
      {
        logit ("et",
               "%s(): Cannot (re)allocate memory for 16->32 bit conversion "
               "buffer\n",
               __func__);
        return -1;
      }
      int32count = msr->numsamples;
    }

    /* Convert 16-bit integers to 32-bit integers */
    int16buffer = (int16_t *)msr->datasamples;

    for (idx = 0; idx < msr->numsamples; idx++)
    {
      int32buffer[idx] = int16buffer[idx];
    }

    msr->datasamples = int32buffer;
    msr->sampletype  = 'i';
  }
  else if (!strcmp (wfhead->datatype, "s4") || !strcmp (wfhead->datatype, "i4"))
    msr->sampletype = 'i';
  else if (!strcmp (wfhead->datatype, "t4") || !strcmp (wfhead->datatype, "f4"))
    msr->sampletype = 'f';
  else if (!strcmp (wfhead->datatype, "t8") || !strcmp (wfhead->datatype, "f8"))
    msr->sampletype = 'd';
  else
  {
    logit ("et", "%s(): Unsupported data type: '%s'\n", __func__,
           wfhead->datatype);
    return -1;
  }

  /* First look for an existing trace structure for this SEED NSLC. */
  mst          = mstg->traces;
  existing_mst = mst_findmatch (mst, 0, msr->network, msr->station,
                                msr->location, msr->channel);

  /* Add data to trace buffer, creating new entry or extending as needed */
  if (!(mst = mst_addmsrtogroup (mstg, msr, 1, -1.0, -1.0)))
  {
    logit ("et", "%s(): Cannot add packet data to trace buffer!\n", __func__);
    return -1;
  }

  /* If we did not create a new MSTrace, then there is no data gap. */
  if (existing_mst == mst)
    existing_mst = NULL;

  /* To keep small variations in the sample rate or time base from accumulating
   * to large errors, re-base the time of the buffer by back projecting from
   * the endtime, which is calculated from the tracebuf starttime and number
   * of samples.  In essence, this maintains a time line based on the starttime
   * of received tracebufs.  It also retains the variations of the sample rate
   * and other characteristics of the original data stream to some degree. */

  mst->starttime =
      mst->endtime -
      (hptime_t)(((double)(mst->numsamples - 1) / mst->samprate * HPTMODULUS) +
                 0.5);
  mst_duration =
      (hptime_t)(((double)(mst->numsamples) / mst->samprate * HPTMODULUS) +
                 0.5);

  /* Allocate & init per-trace stats structure if needed */
  if (!mst->prvtptr)
  {
    if (!(mst->prvtptr = malloc (sizeof (TraceStats))))
    {
      logit ("et", "%s(): Cannot allocate buffer for trace stats!\n", __func__);
      return -1;
    }

    clearmststats (mst);
  }

  ((TraceStats *)mst->prvtptr)->update = dlp_time ();
  ((TraceStats *)mst->prvtptr)->pktcount += 1;

  /* Set flush flag if rate >= flush rate, and, optionally, if duration >= flush
   * duration */
  flushflag =
      (flushfastrate && (mst->samprate >= flushfastrate) &&
       (((flushfastduration > 0) && (mst_duration >= flushfastduration)) ||
        (flushfastduration == 0)));

  /* If there is a gap or overlap between this data record and previously
   * unwritten data for this SEED NSLC, we will have added a new MSTrace
   * to the group. If so, first flush the previously unwritten data. */
  if (existing_mst && existing_mst->numsamples > 0)
  {
    if (verbose >= 2)
    {
      logit ("et",
             "%s(): streamname=%s.%s.%s.%s mst->samprate=%.2lf GAP "
             "flushfast=%d@%" PRId64 " flushflag=%d\n",
             __func__, msr->station, msr->channel, msr->network, msr->location,
             mst->samprate, flushfastrate, MS_HPTIME2EPOCH (flushfastduration),
             1);
    }
    if ((recordspacked = packtrace (existing_mst, 1)) < 0)
    {
      logit ("et", "%s(): Cannot pack trace buffer or send records!\n",
             __func__);
      return -1;
    }
  }

  /* Pack data for this SEED NSLC. */
  if (verbose >= 2)
  {
    logit ("et",
           "%s(): streamname=%s.%s.%s.%s mst->samprate=%.2lf nsamples = %" PRId64
           " flushfast=%d@%" PRId64 " flushflag=%d\n",
           __func__, msr->station, msr->channel, msr->network, msr->location,
           mst->samprate, mst->numsamples, flushfastrate,
           MS_HPTIME2EPOCH (flushfastduration), flushflag);
  }

  if ((recordspacked = packtrace (mst, flushflag)) < 0)
  {
    logit ("et", "%s(): Cannot pack trace buffer or send records!\n", __func__);
    logit ("et", "%s(): %s.%s.%s.%s  ns=%d  dt=%4.1f  type=%s  sampletype: %c\n",
           __func__, wfhead->sta, wfhead->chan, wfhead->net, wfhead->loc,
           wfhead->nsamp, wfhead->samprate, originaldatatype, mst->sampletype);
    return -1;
  }

  /* Only remove an MSTrace from the group/list if we have multiple */
  /* MSTrace objects for the SNCL in the group and the previously */
  /* existing MSTrace has no samples left.  */
  mst     = mstg->traces;
  prevmst = NULL;
  while (mst && existing_mst && existing_mst->numsamples <= 0)
  {
    /* This is the ONLY place we ever unlink and free an MSTrace.
     * This is only done when we have 2 MSTrace entries for the same NSLC,
     * in which case we free the older MSTrace record.
     * Freeing the MSTrace structure will set the pointer to NULL. */
    if (mst == existing_mst)
    {
      if (verbose)
        logmststats (mst);

      if (!prevmst)
        mstg->traces = mst->next;
      else
        prevmst->next = mst->next;

      mst_free (&mst);
      existing_mst = NULL;
    }
    else
    {
      prevmst = mst;
      mst     = mst->next;
    }
  }

  return recordspacked;
} /* End of handletbuf() */

/*********************************************************************
 * sendrecord:
 *
 * Routine called to send a mseed record somewhere
 *	(to a file for now, mseed packets converter later).
 *
 *********************************************************************/
static void
sendrecord (char *record, int reclen, void *handlerdata)
{
  static MSRecord *msr = NULL;
  MSTrace *mst         = handlerdata;
  TraceStats *stats;
  hptime_t endtime;
  char streamid[100];
  int writeack = 0;
  int rv;

  if (!record)
    return;

  /* Parse miniSEED header */
  if ((rv = msr_unpack (record, reclen, &msr, 0, 0)) != MS_NOERROR)
  {
    ms_recsrcname (record, streamid, 0);
    logit ("et", "%s(): Error unpacking %s: %s", __func__, streamid,
           ms_errorstr (rv));
    return;
  }

  /* Generate stream ID for this record: NET_STA_LOC_CHAN/MSEED */
  msr_srcname (msr, streamid, 0);
  strcat (streamid, "/MSEED");

  /* Determine high precision end time */
  endtime = msr_endtime (msr);

  /* Send record to server, loop */
  while (dl_write (dlcp, record, reclen, streamid, msr->starttime, endtime,
                   writeack) < 0)
  {
    if (dlcp->link == -1)
      dl_disconnect (dlcp);

    if (stopsig)
    {
      logit ("et",
             "%s(): Termination signal with no connection to DataLink, the data "
             "buffers will be lost\n",
             __func__);
      stopsig = 2;
      break;
    }

    if (!reconnectinterval)
    {
      stopsig = 2;
      logit ("et",
             "%s(): Termination signal, reconnectinterval=%d, stopsig=%d\n",
             __func__, reconnectinterval, stopsig);
      break;
    }
    else if (dl_connect (dlcp) < 0)
    {
      logit ("et",
             "%s(): Error re-connecting to DataLink server: %s, sleeping\n",
             __func__, dlcp->addr);
      dlp_usleep (reconnectinterval * 1e6);
    }
  }

  /* Update stats */
  if (mst)
  {
    stats = (TraceStats *)mst->prvtptr;

    if (stats->earliest == HPTERROR || stats->earliest > msr->starttime)
      stats->earliest = msr->starttime;

    if (stats->latest == HPTERROR || stats->latest < endtime)
      stats->latest = endtime;

    stats->xmit = dlp_time ();
    stats->reccount += 1;
  }
}

/*********************************************************************
 * packtrace:
 *
 * Package remaining data in buffer for specified MSTrace into
 * miniSEED records.
 *
 * If the flush argument is true the stream buffer will be flushed
 * completely, otherwise records are only packed when enough samples
 * are available to fill a record.
 *
 * Returns the number of records packed on success and -1 on error.
 *********************************************************************/
static int
packtrace (MSTrace *mst, int flush)
{
  static struct blkt_1000_s Blkt1000;
  static struct blkt_1001_s Blkt1001;
  static MSRecord *mstemplate = NULL;

  void *handlerdata = mst;
  int packedrecords = 0;
  int encoding;

  if (!mst)
    return -1;

  /* Set up MSRecord template, include blockette 1000 and 1001 */
  if ((mstemplate = msr_init (mstemplate)) == NULL)
  {
    logit ("et", "%s(): Cannot initialize packing template (out of memory?)\n",
           __func__);
    return -1;
  }

  mstemplate->dataquality = 'D';

  /* Add blockettes 1000 & 1001 to template */
  memset (&Blkt1000, 0, sizeof (struct blkt_1000_s));
  msr_addblockette (mstemplate, (char *)&Blkt1000, sizeof (struct blkt_1001_s),
                    1000, 0);
  memset (&Blkt1001, 0, sizeof (struct blkt_1001_s));
  msr_addblockette (mstemplate, (char *)&Blkt1001, sizeof (struct blkt_1001_s),
                    1001, 0);

  strcpy (mstemplate->network, mst->network);
  strcpy (mstemplate->station, mst->station);
  strcpy (mstemplate->location, mst->location);
  strcpy (mstemplate->channel, mst->channel);

  /* Determine encoding based on sample type */
  if (mst->sampletype == 'f')
    encoding = DE_FLOAT32;
  else if (mst->sampletype == 'd')
    encoding = DE_FLOAT64;
  else
    encoding = int32encoding;

  /* Try Steim-2 first if using dynamic 32-bit integer encoding */
  if (encoding == DYNAMIC_STEIM21I)
    encoding = DE_STEIM2;

  packedrecords = mst_pack (mst, sendrecord, handlerdata, msreclen, encoding, 1,
                            NULL, flush, 0, mstemplate);

  /* Retry with Steim-1 if failed and using dynamic 32-bit integer encoding */
  if (packedrecords == -1 && encoding == DE_STEIM2 &&
      int32encoding == DYNAMIC_STEIM21I)
  {
    logit ("et",
           "%s(): Failed to compress data with Steim-2 %s_%s_%s_%s, trying "
           "Steim-1\n",
           __func__, mst->network, mst->station, mst->location, mst->channel);

    encoding      = DE_STEIM1;
    packedrecords = mst_pack (mst, sendrecord, handlerdata, msreclen, encoding,
                              1, NULL, flush, 0, mstemplate);
  }

  /* Retry with raw 32-bit integers if failed and using dynamic 32-bit integer
   * encoding */
  if (packedrecords == -1 && encoding == DE_STEIM1 &&
      int32encoding == DYNAMIC_STEIM21I)
  {
    logit ("et",
           "%s(), Failed to compress data with Steim-1 %s_%s_%s_%s, trying "
           "32-bit integers\n",
           __func__, mst->network, mst->station, mst->location, mst->channel);

    encoding      = DE_INT32;
    packedrecords = mst_pack (mst, sendrecord, handlerdata, msreclen, encoding,
                              1, NULL, flush, 0, mstemplate);
  }

  return packedrecords;
} /* End of packtrace() */

/*********************************************************************
 * packalltraces:
 *
 * Package data in buffer(s) into miniSEED records.
 *
 * If the flush argument is true the stream buffers will be flushed
 * completely, otherwise records are only packed when CHAD
 *
 * Returns the number of records packed on success and -1 on error.
 *********************************************************************/
static int
packalltraces (int flush, hptime_t flushtime)
{
  MSTrace *mst;

  int trpackedrecords = 0;
  int packedrecords   = 0;
  int flushflag       = flush;

  /* Process all MSTrace entries in the group, flushing if either:
   * a.  the flush flag was specified by the caller, or
   * b.  the trace is older than the flushtime specified by the caller. */
  mst = mstg->traces;

  while (mst && stopsig != 2)
  {
    flushflag = flush;

    if (mst->numsamples > 0)
    {
      /* Determine if last update time for traces is older than flush time */
      if (flushflag == 0 && mst->prvtptr && flushtime != HPTERROR &&
          ((TraceStats *)mst->prvtptr)->update < flushtime)
      {
        if (verbose >= 1)
          logit ("et",
                 "%s(): Flushing data buffer for %s_%s_%s_%s due to flush "
                 "latency\n",
                 __func__, mst->network, mst->station, mst->location,
                 mst->channel);

        flushflag = 1;
      }

      trpackedrecords = packtrace (mst, flushflag);

      if (trpackedrecords == -1)
        return -1;

      packedrecords += trpackedrecords;
    }

    /* Log and clear the channel stats if we flushed this channel. */
    if (verbose && mst->numsamples <= 0 && flushflag)
    {
      logmststats (mst);
      clearmststats (mst);
    }

    mst = mst->next;
  }

  return packedrecords;
} /* End of packalltraces() */

/*********************************************************************
 * logmststats:
 *
 * Log MSTrace stats.
 *********************************************************************/
static void
logmststats (MSTrace *mst)
{
  TraceStats *stats;
  char etime[50];
  char ltime[50];
  char utime[50];
  char xtime[50];

  stats = (TraceStats *)mst->prvtptr;
  ms_hptime2mdtimestr (stats->earliest, etime, 1);
  ms_hptime2mdtimestr (stats->latest, ltime, 1);
  ms_hptime2mdtimestr (stats->update, utime, 1);
  ms_hptime2mdtimestr (stats->xmit, xtime, 1);

  logit ("t", "%s(): %s_%s_%s_%s, earliest: %s, latest: %s\n", __func__,
         mst->network, mst->station, mst->location, mst->channel,
         (stats->earliest == HPTERROR) ? "NONE" : etime,
         (stats->latest == HPTERROR) ? "NONE" : ltime);
  logit ("t", "  last update: %s, xmit time: %s\n",
         (stats->update == HPTERROR) ? "NONE" : utime,
         (stats->xmit == HPTERROR) ? "NONE" : xtime);
  logit ("t", "  pktcount: %lld, reccount: %lld\n",
         (long long int)stats->pktcount, (long long int)stats->reccount);
} /* End of logmststats() */
/*********************************************************************
 * clearmststats:
 *
 * Clear MSTrace stats.
 *********************************************************************/
static void
clearmststats (MSTrace *mst)
{
  if (mst->prvtptr)
  {
    ((TraceStats *)mst->prvtptr)->earliest = HPTERROR;
    ((TraceStats *)mst->prvtptr)->latest   = HPTERROR;
    ((TraceStats *)mst->prvtptr)->update   = HPTERROR;
    ((TraceStats *)mst->prvtptr)->xmit     = HPTERROR;
    ((TraceStats *)mst->prvtptr)->pktcount = 0;
    ((TraceStats *)mst->prvtptr)->reccount = 0;
  }
}

/***************************************************************************
 * logit_msg() and logit_err():
 *
 * Hooks for Earthworm logging facility.  These are used via function
 * pointers by the SeedLink library.
 ***************************************************************************/
static void
logit_msg (const char *msg)
{
  logit ("t", "%s", msg);
}

static void
logit_err (const char *msg)
{
  logit ("et", "%s", msg);
}

#ifndef WIN32
/***************************************************************************
 * term_handler:
 *
 * Signal handler routine to set the termination flag.
 ***************************************************************************/
static void
term_handler (int sig)
{
  logit ("e", "%s()): term_handler caught sig=%d\n", __func__, sig);
  stopsig = 1;
}
#endif
