#include "tbufprocess.h"
#include "swap.h"
#include "trace_buf.h"
#include <string.h>

void
tbufprocess_msr3_free (tbufprocess_input *ti)
{
  if (ti->msr)
  {
    if (!ti->dynamicdatasamples)
    {
      // datasamples was not dynamically allocated
      ti->msr->datasamples = NULL;
    }
    msr3_free (&ti->msr);
  }
}

void
tbufprocess_mstl3_free (tbufprocess_input *ti)
{
  if (ti->mstl)
  {
    mstl3_free (&ti->mstl, 0);
  }
}

void
tbufprocess_free (tbufprocess_input *ti)
{
  if (ti)
  {
    tbufprocess_msr3_free (ti);
    tbufprocess_mstl3_free (ti);
  }
}

int
tbufprocess_init (tbufprocess_input *ti)
{
  memset (ti, 0, sizeof (tbufprocess_input));
  ti->flags |= MSF_PACKVER2;
  ti->encoding = DE_STEIM2;
  ti->reclen   = MS_PACK_DEFAULT_RECLEN;
  if (!(ti->mstl = mstl3_init (ti->mstl)))
  {
    ms_log (2, "%s(): Could not allocate MS3TraceList, out of memory?\n", __func__);
    return -1;
  }
  return 0;
}

void
record_handler (char *msg, int msglength, void *handlerdata)
{
  if (handlerdata)
  {
    tbufprocess_input *ti = (tbufprocess_input *)handlerdata;
    if (ti->verbose)
    {
      ms_log (0, "%s(): msglength=%d\n", __func__, msglength);
    }
    if (ti->mh)
    {
      ti->mh (msg, msglength, tbufprocess_format (ti));
    }
  }
}

// Process a TBUF message and call the specified callback handlers
// ti - tbufprocess input
// msg - the message data or NULL to flush
// returns 0 if successful
int
tbufprocess (tbufprocess_input *ti, char *msg)
{
  int retval     = 0;
  uint32_t flags = ti->flags;
  if (msg)
  {
    TRACE2_HEADER *trh2 = (TRACE2_HEADER *)msg;
    // swap bytes if necessary
    retval = WaveMsg2MakeLocal (trh2);
    if (retval)
    {
      ms_log (2, "Error with WaveMsg2MakeLocal(): %d\n", retval);
      return -1;
    }
    if (!(ti->msr = msr3_init (ti->msr)))
    {
      ms_log (2, "%s(): Could not allocate MS3Record, out of memory?\n", __func__);
      return -1;
    }
    if (trh2->nsamp <= 0)
    {
      if (ti->verbose)
      {
        ms_log (0, "%s(): Empty TBUF message\n", __func__);
      }
      return 0;
    }
    switch (*trh2->datatype)
    {
    case 'f': // INTEL float
      if (trh2->datatype[1] == '8')
      {
        ti->msr->sampletype = 'd';
      }
      else
      {
        ti->msr->sampletype = 'f';
      }
      break;
    case 'i': // INTEL int
      ti->msr->sampletype = 'i';
      break;
    case 's': // SPARC int
      ti->msr->sampletype = 'i';
      break;
    case 't': // SPARC float
      if (trh2->datatype[1] == '8')
      {
        ti->msr->sampletype = 'd';
      }
      else
      {
        ti->msr->sampletype = 'f';
      }
      break;
    default:
      ti->msr->sampletype = 'i';
      break;
    }
    retval = ms_nslc2sid (ti->msr->sid, LM_SIDLEN, 0, trh2->net, trh2->sta, trh2->loc, trh2->chan);
    if (retval < 0)
    {
      ms_log (2, "%s(): Could not determine sid\n", __func__);
      msg = NULL; // ignore message
    }
    else
    {
      ti->msr->starttime   = MS_EPOCH2NSTIME (trh2->starttime);
      ti->msr->encoding    = ti->encoding;
      ti->msr->pubversion  = 1;
      ti->msr->datasamples = (void *)(msg + sizeof (TRACE2_HEADER));
      ti->msr->samplecnt = ti->msr->numsamples = trh2->nsamp;
      ti->msr->samprate                        = trh2->samprate;
      if (ti->verbose)
      {
        ms_log (0, "%s(): nsamp=%d, samprate=%f, datatype=%s, sampletype=%c\n",
                __func__,
                trh2->nsamp, trh2->samprate, trh2->datatype, ti->msr->sampletype);
      }
    }
  }
  if (ti->mstl)
  {
    int64_t psamples = 0;
    if (msg)
    {
      /* Add record to trace list */
      if (mstl3_addmsr (ti->mstl, ti->msr, 0, 1, 0, NULL) == NULL)
      {
        ms_log (2, "%s(): Error adding MS3Record to MS3TraceList\n", __func__);
      }
    }
    else
    {
      flags |= MSF_FLUSHDATA;
    }
    retval = mstl3_pack (ti->mstl,        // Pack data in this trace list
                         &record_handler, // Callback function that will handle records
                         ti,              // Callback function data
                         ti->reclen,      // Maximum record length
                         ti->encoding,    // Data encoding
                         &psamples,       // Packed sample count
                         flags,           // Flags to control packing
                         ti->verbose,     // Verbosity
                         NULL             // Extra headers to inject, none in this case
    );
    if (retval < 0)
    {
      ms_log (2, "%s(): Could not pack MSEED\n", __func__);
      msg = NULL; // ignore message
    }
  }
  else
  {
    int64_t psamples = 0;
    retval           = msr3_pack (ti->msr, &record_handler, ti, &psamples, flags, ti->verbose);
    if (retval < 0)
    {
      ms_log (2, "%s(): Could not pack MSEED\n", __func__);
      msg = NULL; // ignore message
    }
  }
  tbufprocess_msr3_free (ti);
  return retval;
}
